import ast

from types import FrameType
from typing import Any

from .macro import macro, MacroLevel
from ._internal.placeholder import __Placeholder

__all__ = ["f", "_", "_1", "_2", "_3", "_4", "_5", "_6", "_7", "_8", "_9"]


class __f:
    @staticmethod
    @macro(level=MacroLevel.AST)
    def __getitem__(args_src: ast.expr, frame: FrameType):
        placeholders = set()
        for node in ast.walk(args_src):
            if isinstance(node, ast.Name) and node.id in [
                "_",
                *(f"_{i}" for i in range(1, 10)),
            ]:
                placeholders.add(node.id)

        if "_" in placeholders and placeholders - {"_"}:
            raise ValueError(
                "A quick lambda should use either _ or _1, _2, ..., but not both."
            )

        if "_" in placeholders:
            args = ["_"]
        else:
            args = sorted(placeholders, key=lambda x: int(x[1:]))

        args = map(lambda x: ast.arg(arg=x, annotation=None), args)
        ast_func = ast.Lambda(
            args=ast.arguments(
                list(args), [], kwonlyargs=[], kw_defaults=[], defaults=[]
            ),
            body=args_src,
        )
        return eval(
            ast.unparse(ast_func),
            frame.f_globals,
            frame.f_locals,
        )


f = __f()
"""
```py
# one argument
func = f[_ + 1]  # equivalent to: func = lambda _: _ + 1
# multiple arguments
func = f[_1 + _2 + _3]  # equivalent to: func = lambda _1, _2, _3: _1 + _2 + _3
```
"""


_: Any = __Placeholder()
_1, _2, _3, _4, _5, _6, _7, _8, _9 = [_] * 9



from ._internal import create_cate
# fmt: off
from .cates import (
    Semigroup, register_semigroup, sconcat,  # Semigroup
    Monoid, register_monoid, semigroup2monoid, mempty, mappend,  # Monoid
    Functor, register_functor, fmap,  # Functor
    Monad, register_monad, bind, join,  # Monad
)

__all__ = [
    "create_cate",
    "Semigroup", "register_semigroup", "sconcat",  # Semigroup
    "Monoid", "register_monoid", "semigroup2monoid", "mempty", "mappend",   # Monoid
    "Functor", "register_functor", "fmap",  # Functor
    "Monad", "register_monad", "bind", "join",  # Monad
]
# fmt: on



"""
Functional style control flow
"""

from collections.abc import Callable
from typing import TypeVar


__all__ = ["Break", "While"]


class _Break: ...


Break = _Break()


T = TypeVar("T")


def While(func: Callable[[T], T | _Break], initial: T) -> T:
    prev = initial
    while not isinstance(res := func(prev), _Break):
        prev = res
    return prev



import ast
import inspect
import linecache
import enum

from collections.abc import Callable
from typing import TypeVar, cast

from ..utils import pipe
from ..pattern_match import match, matchV, case, default


__all__ = ["MacroLevel", "macro"]


def _get_args_ast(frame_info: inspect.FrameInfo) -> ast.expr | list[ast.expr]:
    """
    Get the source code of the frame caller arguments

    ```py
    stack = inspect.stack()
    frame_info = stack[1]
    src = get_args_src(frame_info)
    ```
    """
    frame = frame_info.frame
    file = inspect.getsourcefile(frame)
    if not file:
        raise ValueError("Could not determine source file.")

    position = frame_info.positions
    if not (position and position.lineno and position.end_lineno):
        raise ValueError("Could not determine source code positions.")

    # get source code
    lines = map(
        lambda line: linecache.getline(file, line),
        range(position.lineno, position.end_lineno + 1),
    )
    lines = list(lines)
    if len(lines) > 1:
        lines[0] = lines[0][position.col_offset :]
        lines[-1] = lines[-1][: position.end_col_offset]
    else:
        # if only one line, slice once since col_offset and end_col_offset are relative to the same line
        lines[0] = lines[0][position.col_offset : position.end_col_offset]
    src = "".join(lines)

    # process AST
    ast_tree = ast.parse(src)
    expr = ast_tree.body[0]

    if not isinstance(expr, ast.Expr):
        raise ValueError("Could not get expression")

    expr = expr.value

    # TODO: currently only support Call and Subscript
    if isinstance(expr, ast.Call):
        return expr.args
    elif isinstance(expr, ast.Subscript):
        return expr.slice
    else:
        raise ValueError("Could not get arguments")


@enum.unique
class MacroLevel(enum.Enum):
    AST = 1
    STR = 2


# TODO: explicit typing for allowed arguments with type parameters list (>= 3.12)
def macro(*, level: MacroLevel = MacroLevel.AST):
    R = TypeVar("R")

    def decorator(func: Callable[..., R]) -> Callable[..., R]:
        def wrapper(*_, **__):
            stack = inspect.stack()
            frame_info = stack[1]
            args_ast = _get_args_ast(frame_info)

            # fmt: off
            args = pipe(
                level,
                match(
                    case(MacroLevel.AST) >> args_ast,
                    case(MacroLevel.STR) >> matchV(args_ast)(
                        case(lambda x: isinstance(x, list)) >> (lambda x: [ast.unparse(arg) for arg in x]),
                        default >> ast.unparse,
                    ),
                )
            )
            # fmt: on
            return func(args, frame_info.frame)

        return cast(Callable[..., R], wrapper)

    return decorator



import operator

from fp_cate.utils import assert_and, flip

from ._internal import create_cate


# fmt: off
__all__ = [
    "Semigroup", "register_semigroup", "sconcat",  # Semigroup
    "Monoid", "register_monoid", "semigroup2monoid", "mempty", "mappend",   # Monoid
    "Functor", "register_functor", "fmap",  # Functor
    "Monad", "register_monad", "bind", "join",  # Monad
]
# fmt: on

Semigroup, register_semigroup, sconcat = create_cate("Semigroup", "sconcat")
"""
```
register_semigroup(type, *, sconcat: Callable) -> None
```

Semigroup operation:
```
sconcat(a: T, b: T) -> T
```
"""

register_semigroup(int, sconcat=operator.add)
register_semigroup(float, sconcat=operator.add)
register_semigroup(str, sconcat=operator.add)
register_semigroup(list, sconcat=operator.add)
register_semigroup(tuple, sconcat=operator.add)
register_semigroup(dict, sconcat=lambda a, b: {**a, **b})
register_semigroup(set, sconcat=operator.or_)


Monoid, register_monoid, mempty, mappend = create_cate("Monoid", "mempty", "mappend")
"""
```
register_monoid(type, *, mempty: Callable, mappend: Callable) -> None
```

Monoid operations:
```
mempty(value: T) -> T
mappend(a: T, b: T) -> T
```
"""
semigroup2monoid = lambda t, *, mempty: assert_and(  # noqa: E731
    issubclass(t, Semigroup),
    register_monoid(t, mempty=mempty, mappend=sconcat),
    msg="Type must be a Semigroup",
)

semigroup2monoid(int, mempty=lambda _: 0)
semigroup2monoid(float, mempty=lambda _: 0.0)
semigroup2monoid(str, mempty=lambda _: "")
semigroup2monoid(list, mempty=lambda _: [])
semigroup2monoid(tuple, mempty=lambda _: ())
semigroup2monoid(dict, mempty=lambda _: {})
semigroup2monoid(set, mempty=lambda _: set())

Functor, _register_functor, _fmap = create_cate("Functor", "fmap")
fmap = flip(_fmap)  # fliping arguments since dispatching is on the first argument
register_functor = lambda t, *, fmap: _register_functor(  # noqa: E731
    t, fmap=flip(fmap)
)
"""
```
register_functor(type, *, fmap: Callable) -> None
```

Functor operation:
```
fmap(func: Callable[[A], B], fa: F[A]) -> F[B]
```
"""

register_functor(list, fmap=lambda f, xs: [f(x) for x in xs])
register_functor(tuple, fmap=lambda f, xs: tuple(f(x) for x in xs))
register_functor(dict, fmap=lambda f, d: {k: f(v) for k, v in d.items()})
register_functor(set, fmap=lambda f, s: {f(x) for x in s})


Monad, register_monad, bind, join = create_cate("Monad", "bind", "join")
"""
Monad operations:
```
bind(ma: M[A], f: Callable[[A], M[B]]) -> M[B]
join(mma: M[M[A]]) -> M[A]
```
"""

register_monad(
    list,
    bind=lambda xs, f: [y for x in xs for y in f(x)],
    join=lambda xss: [y for xs in xss for y in xs],
)
register_monad(
    tuple,
    bind=lambda xs, f: tuple(y for x in xs for y in f(x)),
    join=lambda xss: tuple(y for xs in xss for y in xs),
)
register_monad(
    set,
    bind=lambda xs, f: {y for x in xs for y in f(x)},
    join=lambda xss: {y for xs in xss for y in xs},
)



from .pattern_match import case, matchV, match, switch, _any, _rest, default
from .flow import Break, While

# fmt: off
__all__ = [
    # pattern_match
    "case", "matchV", "match", "switch", "_any", "_rest", "default",
    # flow
    "Break", "While"
]
# fmt: on



"""
Functional style wrappers for list operations that allow for in-place or out-of-place modifications.
"""

from typing import TypeVar

__all__ = ["insert", "remove", "append", "extend", "swap"]


T = TypeVar("T")


_inplace_list = lambda arr, inplace: arr if inplace else arr[:]  # noqa: E731


def insert(arr: list[T], index: int, value: T, inplace=False) -> list[T]:
    arr = _inplace_list(arr, inplace)
    arr.insert(index, value)
    return arr


def remove(arr: list[T], index: int, inplace=False) -> list[T]:
    """Remove the element at the specified index from the list."""
    arr = _inplace_list(arr, inplace)
    arr.pop(index)
    return arr


def append(arr: list[T], value: T, inplace=False) -> list[T]:
    arr = _inplace_list(arr, inplace)
    arr.append(value)
    return arr


def extend(arr: list[T], values: list[T], inplace=False) -> list[T]:
    arr = _inplace_list(arr, inplace)
    arr.extend(values)
    return arr


def swap(arr: list[T], i: int, j: int, inplace=False) -> list[T]:
    """Swap the elements at index i and j in the list."""
    arr = _inplace_list(arr, inplace)
    arr[i], arr[j] = arr[j], arr[i]
    return arr



"""
Example of pattern matching.

```py
# works with any iterables
a = "test"
print(
    matchV(a)(
        case("tes") >> (lambda x: "one"),
        case(["a", _rest]) >> (lambda x, xs: f"list starts with a, rest is {xs}"),
        default >> "good",
    )
)
# works with dicts
pipe(
    {"test": 1, "other": 2},
    match(
        case({"test": _any}) >> (lambda x: f"test is {x}"),
        case({"other": 2}) >> (lambda x: "other two"),
    ),
    print,
)

# works with dataclasses as well
@dataclass
class Test:
    a: int
    b: bool
pipe(
    Test(1, True),
    match(
        case({"a": 1}) >> "this is a good match",
        case({"b": False}) >> "this won't match",
        default >> "all other matches failed",
    ),
    print,
)
```
"""

from collections.abc import Callable, Iterable
from typing import Any

__all__ = ["case", "default", "matchV", "match", "switch", "_any", "_rest"]


# special symbols for pattern matching
_any = lambda *_: True  # noqa: E731
_rest = object()


class _InternalMatchError(Exception): ...


def _get_obj_property(obj: Any, key: str) -> Any:
    try:
        try:
            return obj[key]
        except (KeyError, TypeError):
            return getattr(obj, key)
    except AttributeError:
        raise _InternalMatchError()


class case:
    def __init__(self, pattern: Any, expr: Callable | Any | None = None):
        self._pattern = pattern
        self.expr: Callable = expr if callable(expr) else (lambda *_: expr)

    def __rshift__(self, expr: Callable | Any):
        self.expr = expr if callable(expr) else (lambda *_: expr)
        return self

    def match(self, value: Any) -> tuple | bool:
        return self._match(self._pattern, value)

    def _match(self, pattern: Any, value: Any) -> tuple | bool:
        if isinstance(pattern, dict) and self._match_dict(pattern, value):
            # match itemgetter or object property with dict
            return True
        elif isinstance(pattern, list) and isinstance(value, Iterable):
            # match iterable with list
            if (res := self._match_iter(pattern, value)) is not False:
                return res
        elif callable(pattern):
            # pattern is a predicate function
            res = pattern(value)
            if isinstance(res, bool) and res:
                return True
            elif isinstance(res, tuple):
                return res

        return pattern == value

    def _match_dict(self, pattern: dict, value: Any) -> bool:
        """use dict to match properties of an object"""
        keys = pattern.keys()
        try:
            if all(self._match(pattern[k], _get_obj_property(value, k)) for k in keys):
                return True
        except _InternalMatchError:
            pass

        return False

    def _match_iter(self, pattern: list, value: Iterable) -> tuple | bool:
        """use list to match elements of an iterable"""
        res = ()

        v_list = list(value)

        if len(pattern) > len(v_list):
            # pattern longer than value, directly fail
            return False

        for p, v in zip(pattern, value):
            if p is _rest:
                res += (v_list[len(res) :],)
                return res
            m = self._match(p, v)
            if m is False:
                return False

            # discard nested tuples for simplicity of usage now
            # res += (m,) if isinstance(m, tuple) else (v,)
            res += (v,)

        return res


default = case(_any)


def _match(value: Any, cases: Iterable[case]) -> Any:
    for c in cases:
        if (res := c.match(value)) is not False:
            return c.expr(*res) if isinstance(res, tuple) else c.expr(value)
    raise ValueError("no matching case found")


def matchV(value: Any):
    return lambda *args: _match(value, args)


switch = matchV(...)
"""
Simplified version of matchV for switch-case style usage.
"""


def match(*cases: case) -> Callable[[Any], Any]:
    return lambda v: _match(v, cases)



from collections.abc import Callable
from functools import partial, reduce
from inspect import _empty, signature
from typing import TypeVar, overload

__all__ = [
    "identity",
    "assert_and",
    "expr",
    "compose",
    "curry",
    "pipe",
    "dbg",
    "flip",
    "Filter",
    "State",
]


def identity(x):
    return x


T = TypeVar("T")


def assert_and(cond: bool, expr: T, msg: str | None = "Assertion failed") -> T:
    """Works like assert but returns expr when cond is True"""
    assert cond, msg
    return expr


def expr(*args):
    """Evaluates and returns the last argument"""
    return args[-1]


def compose(*funcs):
    """Composes multiple functions into a single function"""
    return reduce(lambda f, g: lambda x: f(g(x)), funcs, identity)


# from https://stackoverflow.com/a/78149460
# TODO: since it uses inspect, it doesn't work with built-in functions
def curry(f):
    def inner(*args, **kwds):
        new_f = partial(f, *args, **kwds)
        params = signature(new_f, follow_wrapped=True).parameters
        if all(params[p].default != _empty for p in params):
            return new_f()
        else:
            return curry(new_f)

    return inner


def pipe(value, *funcs):
    return reduce(lambda v, f: f(v), funcs, value)


@overload
def dbg(value: T, *, formatter: Callable[[T], str]) -> T: ...


@overload
def dbg(msg: str, value: T, *, formatter: Callable[[T], str]) -> T: ...


def dbg(msg_or_value, value=None, *, formatter=identity):  # type: ignore
    if value is None:
        value = msg_or_value
        print(f"{formatter(value)}")
        return value
    else:
        msg = msg_or_value
        print(f"{msg} {formatter(value)}")
        return value


U = TypeVar("U")
V = TypeVar("V")


def flip(f: Callable[[U, V], T]) -> Callable[[V, U], T]:
    return lambda x, y: f(y, x)


class Filter:
    def __init__(self, func):
        self.func = func

    def __rmatmul__(self, iterable):
        return type(iterable)(filter(self.func, iterable))


class State:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getitem__(self, key):
        return self.__dict__[key]

    def update(self, **kwargs) -> "State":
        new_state = self.__dict__.copy()
        return State(**{**new_state, **kwargs})

    def __repr__(self):
        return (
            "Ctx {\n"
            + "\n".join(f"  {k}={v}" for k, v in self.__dict__.items())
            + "\n}"
        )



from .utils import (
    identity,
    assert_and,
    expr,
    compose,
    curry,
    pipe,
    dbg,
    flip,
    Filter,
    State,
)

# fmt: off
__all__ = [
    # utils
    "identity", "assert_and", "expr", "compose", "curry", "pipe",
    "dbg", "flip",
    "Filter", "State",
]
# fmt: on



"""
Functional style wrappers for dict operations that allow for in-place or out-of-place modifications.
"""

from typing import TypeVar

__all__ = ["insert", "remove", "update"]

K = TypeVar("K")
V = TypeVar("V")

_inplace_dict = lambda d, inplace=False: d if inplace else d.copy()  # noqa: E731


def insert(d: dict[K, V], key: K, value: V, inplace=False) -> dict[K, V]:
    d = _inplace_dict(d, inplace)
    d[key] = value
    return d


def remove(d: dict[K, V], key: K, inplace=False) -> dict[K, V]:
    d = _inplace_dict(d, inplace)
    d.pop(key, None)
    return d


def update(d: dict[K, V], other: dict[K, V], inplace=False) -> dict[K, V]:
    d = _inplace_dict(d, inplace)
    d.update(other)
    return d



from abc import ABC
from collections.abc import Callable
from functools import singledispatch
from typing import NoReturn
from typing_extensions import Unpack

__all__ = [
    "_not_implemented",
    "create_cate",
]


def _not_implemented(cate: str, t: type) -> NoReturn:
    raise NotImplementedError(f"{cate} not implemented for type {t}")


def create_cate(
    cate_name: str, *methods: str
) -> tuple[type, Unpack[tuple[Callable, ...]]]:
    class _Internal(ABC): ...

    _Internal.__name__ = cate_name

    registers = []
    reg_dict = {}
    for mname in methods:
        _cate_func = lambda *args: _not_implemented(mname, type(args[0]))  # noqa: E731
        _cate_func.__name__ = f"_{mname}"
        cate_func = singledispatch(_cate_func)

        registers.append(cate_func)
        reg_dict[mname] = cate_func

    def register_cate(t: type, /, **method_impls: Callable) -> None:
        _Internal.register(t)
        for mname, impl in method_impls.items():
            reg_dict[mname].register(t, impl)

    return (_Internal, register_cate, *registers)



__all__ = ["TailCall", "tco"]


class TailCall:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


def tco(func):
    """
    Tail Call Optimization decorator.
    Makes a function support tail call optimization by using the `TailCall` class.

    Example:
    ```py
    fib = tco(lambda n, acc=1: 1 if n <= 2 else TailCall(n - 1, acc + n))
    ```
    """

    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        while isinstance(result, TailCall):
            result = func(*result.args, **result.kwargs)
        return result

    return wrapper



__all__ = ["__Placeholder"]


class __Placeholder:
    def __add__(self, _):
        return self

    def __sub__(self, _):
        return self

    def __mul__(self, _):
        return self

    def __truediv__(self, _):
        return self

    def __floordiv__(self, _):
        return self

    def __mod__(self, _):
        return self

    def __pow__(self, _):
        return self

    def __radd__(self, _):
        return self

    def __rsub__(self, _):
        return self

    def __rmul__(self, _):
        return self

    def __rtruediv__(self, _):
        return self

    def __rfloordiv__(self, _):
        return self

    def __rmod__(self, _):
        return self

    def __rpow__(self, _):
        return self

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    def __not__(self):
        return self

    def __abs__(self):
        return self

    def __invert__(self):
        return self

    def __eq__(self, _):
        return False

    def __ne__(self, _):
        return True

    def __lt__(self, _):
        return False

    def __le__(self, _):
        return False

    def __gt__(self, _):
        return False

    def __ge__(self, _):
        return False

    def __bool__(self):
        return False

    def __call__(self, *_, **__):
        return self

    def __getitem__(self, _):
        return self

    def __repr__(self):
        return "__Placeholder()"

    def __str__(self):
        return "Placeholder"
